# gallenbacher_dijkstra_objektorientiertv.py
# Programm entsprechend der Algorithmusformulierung von Gallenbacher
from gallenbacher_graph import gallenbacherGraph

start='I'
ziel='O'

class Kante():

    kantenListe=[]

    @staticmethod
    def erzeugeKantenListe(graph):
        """graph ist dictionary, nach dem Muster
        {'A' : [['B', 69], ['D', 36], ['M', 36], ['N', 22]], ...}"""
        kanten=[]
        for knoten in graph:
            naechste=graph[knoten]
            for nach in naechste:                
                kanten.append(Kante(knoten,nach[0],nach[1]))
        Kante.kantenListe=kanten
            
    def __init__(self, von, nach, bewertung):
       self.von,self.nach,self.bewertung = von,nach,bewertung

    def gibVon(self): return self.von    
    def gibNach(self): return self.nach
    def gibBewertung(self): return self.bewertung
    def gibDaten(self): return self.von,self.nach,self.bewertung

class Weg():
    def __init__(self,knotenListe,start,ziel,laenge):
        self.knotenListe=knotenListe
        self.start=start
        self.ziel=ziel
        self.laenge=laenge

    def gibStart(self): return self.start
    def gibZiel(self): return self.ziel
    def gibLaenge(self): return self.laenge

    def gibWegDaten(self):
        return self.knotenListe,self.start,self.ziel,self.laenge
    
    def zeigeWeg(self):
        text=str(self.knotenListe)
        text+='\nZiel: '+self.ziel+'l='+str(self.laenge)+' : '
        return text

    def letzter(self):
        return self.knotenListe[len(self.knotenListe)-1]
    
    def nachfolgeKanten(self):
        aktuell=self.letzter()
        nachfolger=[]
        for kante in Kante.kantenListe:
            if aktuell==kante.gibVon():
                nachfolger.append(kante)
        return nachfolger

    def fortsetzenMit(self, kante):
        aktuell=self.letzter()
        if aktuell!=kante.gibVon(): # setzt nicht fort
            return
        else:           # Weg fortsetzen
            self.knotenListe.append(kante.gibNach())
            self.laenge+=kante.gibBewertung()

    def neueWege(self):
        '''erzeugt alle fortsetzenden Wege'''
        neue=[]
        knotenListe,start,ziel,laenge=self.gibWegDaten()        
        for kante in self.nachfolgeKanten():
            liste=[]+knotenListe
            weg=Weg(liste,start,ziel,laenge)
            weg.fortsetzenMit(kante) # der neu erzeugte, nicht self
            neue.append(weg)
        return neue

class prioSchlange():
    def __init__(self, startWeg):
        self.wege=[startWeg]

    def gibWege(self):
        return self.wege

    def hatWege(self):
        return len(self.wege)>0

    def vor(self, erster, zweiter):
        return erster.gibLaenge()<zweiter.gibLaenge()
        
    def holeErsten(self):
        if not self.hatWege(): return None
        return self.wege[0]

    def entferneErsten(self):
        if not self.hatWege(): return None
        self.wege=self.wege[1:]
        
    def fuegeEinenEin(self, neu):
        '''fuegt einen neuen Weg in die PrioWS ein'''
        wege=[]+self.wege   ## echte Kopie der Wege
        pos=len(wege) # ggf hinten ran
        for weg in wege:
            if self.vor(neu,weg):
                pos=wege.index(weg)
                break
        self.wege.insert(pos,neu)

    def fuegeEin(self, wege):
        for weg in wege:
            self.fuegeEinenEin(weg)

    def entferneLaengere(self, aktuell):
        '''entfernt alle Wege zum aktuellen Knoten
        bis auf den kürzesten'''
        kuerzster=None
        wege=[]+self.wege   ## echte Kopie der Wege
        for weg in wege:    ## kuerzesten suchen
            if weg.gibZiel()==aktuell:
                if not kuerzester:
                    kuerzester=weg
                elif self.vor(weg,kuerzester):
                    kuerzester=weg
        for i in range(len(wege)):  ## andere entfernen
            if wege[i].gibZiel()==aktuell:
                if wege[i]!=kuerzester:
                    self.wege.remove(wege[i])

    def zeigeSchlange(self):
        for weg in self.wege:
            print(weg.gibWegDaten())

    def amZiel(self):
        if not self.hatWege(): return False
        aktuell=self.holeErsten()
        if aktuell.letzter()==aktuell.gibZiel():
            return True
        return False

def dijkstra(start,ziel):
    weg=Weg([start],start,ziel,0)
    prioWS=prioSchlange(weg)
    while ( prioWS.hatWege() ) and not (prioWS.amZiel()):
        aktuell=prioWS.holeErsten()
        prioWS.entferneErsten()
        neue=aktuell.neueWege()
        prioWS.fuegeEin(neue)
        ##prioWS.zeigeSchlange()
    if prioWS.amZiel():
        return prioWS.holeErsten().gibWegDaten()
    else:
        return False
    
Kante.erzeugeKantenListe(gallenbacherGraph)
print(dijkstra('I','O'))
##print(dijkstra('I','Q'))

### Tests
### Konvertieren in eine Tupel-Liste:
##kantenTupelListe=[]
##for kante in Kante.kantenListe:
##    kantenTupelListe.append(kante.gibDaten())
    
### initialiseren
##def dijkstraStart(startKnoten,ziel):
##    weg=Weg([start],start,ziel,0)
##    prioWS=prioSchlange(weg)
##    prioWS.entferneErsten()
##    neue=weg.neueWege()
##    prioWS.fuegeEin(neue)
##    prioWS.zeigeSchlange()
    
##weg=Weg(['I'],'I','O',0)
##neu=Weg(['I','B'],'I','O',34)
##print(weg.gibWegDaten())
##print(neu.gibWegDaten())
##print(neu.gibLaenge()<weg.gibLaenge())
